#define _CRT_SECURE_NO_WARNINGS

#define DUMP

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#ifdef WIN32
#include <process.h>
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define lseek _lseek
#define close _close

#define O_BINARY _O_BINARY
#define ctime _ctime
#else
#include <unistd.h>
#include <stdlib.h>

#define O_BINARY 0
#endif

#include "xmalloc.h"
#include "object.h"
#include "util.h"
#include "Attribute.h"
#include "AttributeCache.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "EntityInstance.h"
#include "RelationInstance.h"
#include "LandscapeClassObject.h"
#include "EntityClass.h"
#include "RelationClass.h"
#include "TAInput.h"
#include "boundaries.h"

static CTAInput 		ta;
static CEntityInstance	*g_headP = 0;
static int				g_lines_code     = 0;
static int				g_dead_functions = 0;
static int				g_stripped       = 0;
static int				g_dead_lines     = 0;
static int				g_changed        = 0;
static int				g_skipped        = 0;
static int				g_skipped_files  = 0;
static int				g_failed         = 0;
static int				g_updated        = 0;
int						g_backup         = 0;

int						g_trace_flag     = 0;
int						g_tree_flag      = 0;
int						g_debug_flag     = 0;
int						g_called_from    = 0;
int						g_calls_to       = 0;

static void visit(CEntityInstance *eP, int depth);

/*
static void
trap()
{
}
*/

/* For any call from f(x) to a virtual function A::g(), add as needed
 * presumed calls from f(x) to every virtual function X::g() for which
 * X is a direct subclass of A.
 */

static void
handle_polymorphism(CEntityInstance *function1P, int depth)
{
	CEntityClass		*ecP;
	CRelationClass		*rcP;
	const char			*idP, *mangle1P, *class1P, *mangle2P, *P1;
	CEntityInstance		*vtable1P, *typeinfo1P, *typeinfo2P, *vtable2P;
	CEntityInstance		*function2P;
	CRelationInstance	*riP, *ri1P;
	int					lth, c;

	ecP    = function1P->getEntityClass();
	idP = ecP->getId();
	if (!idP || strcmp(idP, "VF")) {
		// Not a virtual function
		return;
	}

	// Check to see if it is ever called virtually

	for (riP = function1P->m_dstRelListP; ; riP = riP->m_next_dstP) {
		if (!riP) {
			// No virtual calls to this function found
			return;
		}
		rcP = riP->getRelationClass();
		if (!strcmp(idP, "CVF")) {
			// There is a virtual call to this function
			break;
	}	}

	mangle1P = function1P->getAttribute("mangle");
	assert(mangle1P);
	// Find the class component of the mangled function name
	for (class1P = mangle1P; (c = *class1P) && (c < '0' || c > '9'); ++class1P);
	lth = 0;
	for (P1 = class1P; (c = *P1) >= '0' && c <= '9'; ++P1) {
		lth = lth * 10 + c - '0';
	}
	assert(lth);
	lth += P1 - class1P;

	// Find the vtable for this functions class
	// Note a function can be called from many vtables

	for (riP = function1P->m_dstRelListP; ; riP = riP->m_next_dstP) {
		if (!riP) {
			assert(riP);
			return;
		}
		vtable1P = riP->m_srcP;
		ecP      = vtable1P->getEntityClass();
		idP      = ecP->getId();
		if (strcmp(idP, "WV")) {
			// Not a weak variable
			continue;
		}
		mangle2P = vtable1P->getAttribute("mangle");
		if (strncmp(mangle2P, "_ZTV", 4)) {
			// Not a vtable
			continue;
		}
		mangle2P += 4;
		if (!strncmp(mangle2P, class1P, lth)) {
			// Found correct class
			break;
	}	}

	// Find the typeinfo of vtable1

	for (riP = vtable1P->m_srcRelListP; ; riP = riP->m_next_srcP) {
		if (!riP) {
			assert(riP);
			return;
		}
		typeinfo1P = riP->m_dstP;
		ecP        = typeinfo1P->getEntityClass();
		idP        = ecP->getId();
		if (!strcmp(idP, "TI")) {
			break;
	}	}

	mangle1P = 0;

	// For each typeinfo2P which references our typeinfo1P
	// implying that its class directly inherits from class1P

	for (riP = typeinfo1P->m_dstRelListP; riP; riP = riP->m_next_dstP) {
		typeinfo2P = riP->m_srcP;
		ecP        = typeinfo2P->getEntityClass();
		idP        = ecP->getId();
		if (strcmp(idP, "TI")) {
			continue;
		}

		// Find the vtable for typeinfo2P

		for (ri1P = typeinfo2P->m_dstRelListP; ; ri1P = ri1P->m_next_dstP) {
			if (!ri1P) {
				assert(0);
				goto skip_match;
			}
			vtable2P = ri1P->m_srcP;
			ecP      = vtable2P->getEntityClass();
			idP   = ecP->getId();
			if (!strcmp(idP, "TV")) {
				break;
		}	}

		// Find a function in that vtable with the same signature as ours

		for (ri1P = vtable2P->m_srcRelListP; ; ri1P = ri1P->m_next_srcP) {
			function2P = ri1P->m_dstP;
			if (function2P->m_visited) {
				continue;
			}
			ecP        = function2P->getEntityClass();
			idP     = ecP->getId();
			if (!strcmp(idP, "VF")) {
				continue;
			}
			if (!mangle1P) {
				mangle1P = function1P->getAttribute("mangle");
				if (!mangle1P) {
					assert(mangle1P);
					break;
				}
				if (!strncmp(mangle1P, "_ZN", 3)) {
					assert(0);
					break;
				}
				mangle1P += 3;
				lth       = 0;
				for (lth = 0; (c = *mangle1P) >= '0' && c <= '9'; ++mangle1P) {
					lth = lth * 10 + c - '0';
				}
				if (!lth) {
					assert(lth);
					break;
				}
				// Skip the class name component of the mangle
				mangle1P += lth;
			}
			mangle2P = function2P->getAttribute("mangle");
			if (!mangle2P) {
				assert(mangle2P);
				continue;
			}
			if (!strncmp(mangle2P, "_ZN", 3)) {
				assert(0);
				break;
			}
			mangle2P += 3;
			lth       = 0;
			for (lth = 0; (c = *mangle2P) >= '0' && c <= '9'; ++mangle2P) {
				lth = lth * 10 + c - '0';
			}
			if (!lth) {
				assert(lth);
				continue;
			}
			// Skip the class name component of the mangle
			mangle2P += lth;

			if (!strcmp(mangle1P, mangle2P)) {
				visit(function2P, depth+1);
		}	}
skip_match:
		continue;
	}
}

static int
same(CEntityInstance *e1P, CEntityInstance *e2P)
{
	CEntityClass	*ec1P, *ec2P;

	ec1P = e1P->getEntityClass();
	ec2P = e2P->getEntityClass();
	if (ec1P != ec2P) {
		return 0;
	}
	if (strcmp(e1P->getLabel(), e2P->getLabel())) {
		return 0;
	}
	return 1;
}

/* Reduce all references to a function (in a header file) seen in amy
 * source files to a single reference. Don't want to treat a function
 * in a header as dead code unless there really are no links to it
 * from any source file
 */

static void
consolidate(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP, **ePP, *e1P;
	const char			*labelP;
	CRelationInstance	*riP, *nextP;
	const char			*fullpathP;
	char				*P;
	int					diff;

	// Sort by source file name and then by lineno
	for (ecP = ta.getEntityClass(0); ecP; ecP = ecP->m_nextP) {
		labelP = ecP->getId();
		if (!labelP) {
			continue;
		}
		if (strcmp(labelP, "GF") && strcmp(labelP, "LF")) {
			continue;
		}

		for (eP = ecP->m_entityInstancesP; eP; eP = eP->m_nextInClassP) {
			fullpathP = eP->m_fullpathP;
			if (!fullpathP || !strcmp(fullpathP, ".")) {
				P = 0;
				eP->buildPath(&P, 0);
				P = (char *) Xmalloc(P - ((char *) 0) + 1);
				eP->m_fullpathP = P;
				eP->buildPath(&P, 1);
			}
			fullpathP = eP->m_fullpathP;

			for (ePP = &g_headP; (e1P = *ePP); ePP = &(e1P->m_nextInOrderP)) {
				diff = strcmp(fullpathP, e1P->m_fullpathP);
				if (diff < 0) {
					goto found;
				}
				if (diff) {
					continue;
				}
				if (eP->m_lineno < e1P->m_lineno) {
					goto found;
				}
				if (eP->m_lineno == e1P->m_lineno && same(eP, e1P)) {
					// if source is compiled more than once
					// The same function might be named multiple times
					//printf("Consolidating %d:[%s]%s/[%s]%s in %s\n", eP->m_lineno, eP->getId(), eP->getLabel(), e1P->getId(), e1P->getLabel(), eP->m_fullpathP);
					for (riP = eP->m_srcRelListP; riP; riP = nextP) {
						nextP = riP->m_next_srcP;
						// Move edge to be sourced by earlier node
						riP->m_next_srcP   = e1P->m_srcRelListP;
						e1P->m_srcRelListP = riP;
						riP->m_srcP        = e1P;
					}
					for (riP = eP->m_dstRelListP; riP; riP = nextP) {
						nextP = riP->m_next_dstP;
						// Moce edge to target earlier node
						riP->m_next_dstP   = e1P->m_dstRelListP;
						e1P->m_dstRelListP = riP;
						riP->m_dstP        = e1P;
					}
					eP->m_srcRelListP = 0;
					eP->m_dstRelListP = 0;
					eP->m_visited     = SkippedX;
					goto skip;
			}	}	
found:		eP->m_nextInOrderP = *ePP;
			*ePP               = eP;
skip: 		continue;
	}	}	
}

static void
visit(CEntityInstance *eP, int depth)
{
	CRelationInstance	*riP;
	CRelationClass		*rcP;
	const char			*labelP;
	CEntityInstance		*dstP;

	assert(!eP->m_visited);
	eP->m_visited = CalledX;

	if (g_tree_flag) {
		fprintf(stderr, "%p ", eP);
		for (int i = 0; i < depth; ++i) {
			fputs("  ", stderr);
		}
		fprintf(stderr, "%s %d:%s\n", eP->getLabel(), eP->m_lineno, eP->m_fullpathP);
	}

	for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
		rcP    = riP->getRelationClass();
		labelP = rcP->getId();
		dstP   = riP->m_dstP;
		if (labelP[1] == 'E') {
			// External variable or function
			continue;
		}
		switch (labelP[0]) {
		case 'C':
			if (!strcmp(labelP, "CIF")) {
				dstP = dstP->getParent();
			} 
			break;
		} 
		if (!dstP->m_visited) {
			visit(dstP, depth+1);
	}	}

	handle_polymorphism(eP, depth);
}

static void
depth_search(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP;
	const char			*labelP;

	// Find the functions and variables not visited
	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
/*
		if (!strcmp(eP->getId(), "F0xec1ac0")) {
			trap();
		}
*/
		labelP = eP->getLabel();

		/* Some functions called during initialistation */
		if (*labelP == '"') {
			++labelP;
		}
		if (!strncmp(labelP, "global ", 7) ||
			!strncmp(labelP, "__static_", 9)) {
			goto descend;
		}
		if (strcmp(labelP, "main") && strcmp(labelP, "wmain")) {
			continue;
		}
		ecP    = eP->getEntityClass();
		labelP = ecP->getId();
		if (!labelP) {
			continue;
		}
		if (strcmp(labelP, "GF")) {
			continue;
		}
descend:
		if (!eP->m_visited) {
			visit(eP, 0);
}	}	}

#ifndef SSIZE_MAX
#define SSIZE_MAX 0x7FFFFFFF
#endif

typedef enum {
	No_report      = 0,
	Report         = 1,
	Report_verbose = 2
} reportE;

typedef enum {
	No_source     = 0,
	Read_source   = 1,
    Strip_source  = 2,
	Write_source  = 3,
	Update_source = 4
} modeE;

static reportE	g_report = Report;
static modeE    g_mode   = No_source;

const char	*g_filenameP = "";
static char	*g_sourceP   = 0;
static int	g_source_lth = 0;

static void
strip_source(void)
{
	char 	*startP, *atP, *at1P;
	char	c;
	int		stripped;

	stripped = g_stripped;
	for (atP = at1P = startP = g_sourceP; ; ++atP) {
		switch ((c = *atP)) {
		case 0:
			goto done;
		case '#':
			if (!strncmp(atP, "#if !defined(_DEAD_)", 20)) {
				++g_stripped;
				for (;; ++atP) {
					if (!*atP) {
						goto done;
					}
					if (!strncmp(atP, "#endif", 6)) {
						break;
				}	}
				atP = strchr(atP, '\n');
				if (!atP) {
					goto done;
				}
				continue;
			}
			if (!strncmp(atP, "#ifndef _DEAD_", 13) ||
				!strncmp(atP, "#endif /* _DEAD_", 16)) {
				++g_stripped;
				if (at1P > startP && at1P[-1] == '\n') {
					--at1P;
				}
				atP = strchr(atP, '\n');
				if (!atP) {
					goto done;
				}
				continue;
		}	}
		*at1P++ = c;
	}
done:
	*at1P = 0;
	if (g_stripped != stripped) {
		g_changed = 1;
	}
}

static const char	*g_underpathP   = 0;
static int			g_underpath_lth = 0;

static int
load_source(void)
{
	struct stat buf;
	int			ret, lth, segment;
	char		*atP, *endP;
	int			fileno, mode;

	if (g_mode != Update_source) {
		mode = O_RDONLY;
	} else {
		mode = O_RDWR;
	}

	fileno = open(g_filenameP, mode | O_BINARY, 0);
	if (fileno < 0) {
		fprintf(stderr, "** Can't open %s for reading", g_filenameP);
		if (mode == O_RDWR) {
			fprintf(stderr, " and update");
		}
		fprintf(stderr, "\n");
		perror("");
		return -1;
	}
	// We have to scan the input more than once so read the entire input
	// into memory

	ret = fstat(fileno, &buf);
	if (ret != 0) {
		fprintf(stderr,"** fstat %s returned %d\n", g_filenameP, ret);
		perror("");
		goto bad_source;
	}

	// Allocate a large enough memory buffer

	lth = buf.st_size;
	if (!lth) {
		fprintf(stderr, "** %s is empty\n", g_filenameP);
		goto bad_source;
	}

	if (g_source_lth <= lth) {
		Xfree(g_sourceP);
		g_sourceP     = (char *) Xmalloc(lth + 1);
		g_source_lth  = lth;
	}
	atP          = g_sourceP;
	
	// Read the source into memory

	for (endP = atP + lth; atP < endP; atP += ret) {
		segment = endP - atP;
		if (segment > SSIZE_MAX) {
			segment = SSIZE_MAX;
		}
		ret = read(fileno, atP, segment);
		if (ret <= 0) {
			fprintf(stderr, "** read %s returned %d\n", g_filenameP, ret);
			perror("");
bad_source:
			close(fileno);
			return -1;
	}	}
	assert(atP == endP);
	*endP = 0;
	
	close(fileno);

	if (g_mode == Strip_source) {
		if (!g_underpathP || !strncmp(g_underpathP, g_filenameP, g_underpath_lth)) {
			strip_source();
	}	}

	return(0);
}

extern boundaryT	*g_boundariesP;
extern int			g_boundaries;

static void
update_source(void)
{
	char		filename[4096];
	const char	*filenameP;
	FILE		*F;
	int			c;
	char		*atP;
	int			lineno, character, comment;
	boundaryT	*boundaryP, *end_boundaryP;
	int			find_lineno, find_char, add_end;

	if (!g_changed) {
		return;
	}

	++g_updated;
	if (g_mode == Write_source) {
		sprintf(filename, "%s.dead", g_filenameP);
		filenameP = filename;
	} else {
		filenameP = g_filenameP;
		if (g_backup) {
			char backup[4096];
			int	 ret;

			sprintf(backup, "cp %s %s.bak", filenameP, filenameP);
			ret = system(backup);
			if (ret) {
				fprintf(stderr, "%s returned %d\n", backup, ret);
		}	}
	}
	F = fopen(filenameP, "wb");
	if (!F) {
		fprintf(stderr, "%s Can't write\n", filenameP);
		perror("");
		return;
	}
	rewind(F);

	atP = g_sourceP;

	if (g_mode > Read_source) {
		extern	char **g_argsPP;
		char		**argPP;
		char		*argP, *P, *nameP;

		time_t		time_now;
		char		*dateP;
		struct tm 	*tmP;
		char		timestamp[20];

 		time( &time_now );
 		tmP = gmtime(&time_now);
		sprintf(timestamp, "%04d%02d%02d%02d%02d%02d", 		
			tmP->tm_year+1900, tmP->tm_mon+1, tmP->tm_mday,
			tmP->tm_hour, tmP->tm_min, tmP->tm_sec);

		dateP = ctime( &time_now ); 
		dateP[24] = 0;

		fprintf(F, "#if !defined(_DEAD_) ");
		for (argPP = g_argsPP; (argP = *argPP); ++argPP) {
			if (argP[0] == '-' && (argP[1] == 'D' || argP[1] == 'U')) {
				fprintf(F, "&& ");
				nameP = argP + 2;
				P = strchr(nameP, '=');
				if (P) {
					*P = 0;
				}
				if (!P || P[1] == '"') {
					// Can't use strings in the conditional language
					// supported by CPP
					if (argP[1] == 'U') {
						fputc('!', F);
					}
					fprintf(F, "defined(%s) ", nameP);
				} else {
					fprintf(F, "defined(%s) && (%s == %s) ", nameP, nameP, P+1);
				}
				if (P) {
					*P = '=';
				}
		}	}
		fprintf(F, "\n#define _DEAD_%s\n#endif /* _DEAD_ */\n", timestamp);

		lineno        = 1;
		character     = 0;
		comment       = 0;
		boundaryP 	  = g_boundariesP;
		for (end_boundaryP = boundaryP + g_boundaries; boundaryP < end_boundaryP; ++boundaryP) {
			if (!boundaryP->dead || boundaryP->lineno[1] <= 0) {
				continue;
			}
			for (add_end = 0; add_end < 2; ++add_end) {
				find_lineno = boundaryP->lineno[add_end];
				find_char   = boundaryP->character[add_end];
				if (lineno > find_lineno) {
					goto skip;
				}
				for (; (c = *atP); ) {
					switch (comment) {
					case 0:
						if (c == '/') {
							if (atP[1] == '*') {
								fputc(c, F);
								c       = *(++atP);
								comment = 1;
								break;
							}
							if (atP[1] == '/') {
								comment = 2;
								break;
						}	}
						++character;
						break;
					case 1:
						if (c == '*' && atP[1] == '/') {
							fputs("*/", F);
							atP    += 2;
							comment = 0;
							continue;
						}
						break;
					case 2:
						if (c == '\n') {
							comment = 0;
						}
						break;
					}
							
					if (lineno == find_lineno && character == find_char) {
						if (!add_end) {
							fprintf(F, "\n#ifndef _DEAD_%s /* %s */\n", timestamp, dateP);
						} else {
							fprintf(F, "\n#endif /* _DEAD_%s %s */\n", timestamp, dateP);
							g_lines_code += find_lineno - boundaryP->lineno[0] + 1;
						}
						--character;
						break;
					}
					fputc(c, F);
					++atP;
					if (c == '\n') {
						++lineno;
						if (lineno > find_lineno) {
							goto skip;
						}
						character = 0;
			}	}	}
skip:		continue;
	}	}

	for (; (c = *atP); ++atP) {
		fputc(c, F);
	}
	fclose(F);
}

static void
show_callers(CEntityInstance *eP)
{
	CRelationInstance	*riP, *ri1P;
	CEntityInstance		*callerP;
	CEntityClass		*ecP;

	if (g_called_from) {
		for (riP = eP->m_dstRelListP; riP; riP = riP->m_next_dstP) {
			callerP = riP->m_srcP;
			ecP     = callerP->getEntityClass();
			printf("\t\t<- (%s) %s", ecP->getLabel(), callerP->getLabel());
			if (callerP->m_visited) {
				printf(" **VISIted**");	
			}
			printf("\n");
		}
		for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
			callerP = riP->m_dstP;
			ecP     = callerP->getEntityClass();
			if (!strcmp(ecP->getId(),"FA")) {
				for (ri1P = callerP->m_dstRelListP; ri1P; ri1P = ri1P->m_next_dstP) {
					callerP = riP->m_srcP;
					if (callerP == eP) {
						continue;
					}
					ecP     = callerP->getEntityClass();
					printf("\t\t<-&(%s) %s", ecP->getLabel(), callerP->getLabel());
					if (callerP->m_visited) {
						printf(" **VISIted**");	
					}
					printf("\n");
	}	}	}	}
	if (g_calls_to) {
		for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
			callerP = riP->m_dstP;
			ecP     = callerP->getEntityClass();
			printf("\t\t-> (%s) %s", ecP->getLabel(), callerP->getLabel());
			if (callerP->m_visited) {
				printf(" **VISIted**");	
			}
			printf("\n");
	}	}
}

static void
report_dead(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP;
	const char			*labelP;
	const char			*typeP;
	int					load_failed = 1;
	int					boundaries  = -1;
	int					skip        = 0;
	int					lineno;
	boundaryT			*boundaryP, *end_boundaryP;
	const char			*msgP;

	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
		
		if (eP->m_visited) {
			continue;
		}
		if (strcmp(g_filenameP, eP->m_fullpathP)) {
			/* Commit update to last source */

			update_source();
			g_changed   = 0;
			g_filenameP = eP->m_fullpathP;
			load_failed = 1;
			boundaries  = -1;
			skip        = (g_underpathP && strncmp(g_underpathP, g_filenameP, g_underpath_lth));
			if (skip) {
				++g_skipped_files;
				if (g_report == Report_verbose) {
					msgP = "[skipped]";
				} else {
					msgP = 0;
				}
			} else {
				msgP = "";
				if (g_mode != No_source) {
					load_failed = load_source();
					if (load_failed) {
						msgP = " [can't read]";
			}	}	}
			if (g_report != No_report) {
				if (msgP) {
					printf("%s %s\n", g_filenameP, msgP);
		}	}	}

		lineno = eP->m_lineno;
		if (skip) {
			++g_skipped;
			if (g_report == Report_verbose) {
				msgP = "skipped";
				goto report;
			}
			continue;
		}

		++g_dead_functions;
		if (load_failed) {
			msgP = "can't read";
		} else if (lineno == NoLineno) {
			msgP = "include file";
		} else if (g_mode < Write_source) {
			msgP = "dead code";
		} else {
			if (g_mode == Write_source) {
				msgP = "wrote";
			} else {
				msgP = "update";
			}

			if (boundaries == -1) {
				boundaries = establish_boundaries(eP);
			}

			boundaryP = g_boundariesP;
			for (end_boundaryP = boundaryP + boundaries; ; ++boundaryP) {
				if (end_boundaryP <= boundaryP) {
					msgP = "scan failed";
					g_failed++;	
					goto report;
				} 
				if (boundaryP->lineno[0] <= lineno && boundaryP->lineno[1] >= lineno) {
					boundaryP->dead = 1;
					g_changed       = 1;
					break;
		}	}	}

report:
		if (g_report != No_report) {
			ecP    = (CEntityClass *) eP->getParentClass();
			typeP  = ecP->getLabel();
			labelP = eP->getLabel();
			printf("\t(%s) %s", typeP, labelP);
			if (lineno != NoLineno) {
				printf(":%d", lineno);
			}
			printf(" %s\n", msgP);

			if (g_called_from || g_calls_to) {
				show_callers(eP);
			}
	}	}

	update_source();
}

static void
find_dead(const char *filenameP)
{
	ta.loadTA(filenameP);
	consolidate();
	depth_search();
	report_dead();
}

int 
main(int argc, const char *argv[])
{
	const char *filenameP = 0;
	int			i;
	const char	*argP;
	char		cwd[2024];

	printf("%s " VERSION " " __DATE__ " " __TIME__ "\n", argv[0]);

	for (i = 1; i < argc; ++i) {
		argP = argv[i];
		switch (argP[0]) {
		case '-':
			switch (argP[1]) {
			case 'b':
				g_backup = 1;
				break;
			case 'd':
				g_debug_flag = 1;
				break;
			case 'f':
				g_called_from = 1;
				break;
			case 'F':
				g_calls_to = 1;
				break;
			case 'r':
				g_mode = Read_source;
				break;
			case 'p':
				if (argP[2]) {
					g_underpathP = argP+2;
				} else {
					g_underpathP = argv[++i];
					if (!g_underpathP) {
						--i;
				}	}
				if (g_underpathP && !strcmp(g_underpathP, ".")) {
					g_underpathP = getcwd(cwd, sizeof(cwd));
					if (!g_underpathP) {
						fprintf(stderr, "Can't identify CWD\n");
						exit(1);
				}	}
				break;
			case 'w':
				g_mode = Write_source;
				break;
			case 'v':
				g_report = Report_verbose;
				break;
			case 'u':
				g_mode = Update_source;
				break;
			case 's':
				g_mode = Strip_source;
				break;
			case 't':
				g_trace_flag = 1;
				break;
			case 'T':
				g_tree_flag = 1;
				break;
			default:
				fprintf(stderr, "Unknown option %s\n", argP);
				fprintf(stderr, "usage: %s [-r/ead] [-s/trip] [[-b/ackup] -w/rite|-u/pdate] [-v/erbose] [-t/race] [-T/ree] [<] <tafile>\n", argv[0]);
				return(1);
			}
			break;
		default:
			if (filenameP) {
				fprintf(stderr, "Multiple ta_files\n");
				fprintf(stderr, "usage: %s [<] <tafile>\n", argv[0]);
				return(1);
			}
			filenameP = argP;
			break;
	}	}
	if (g_mode == No_source && g_report == No_report) {
		g_report = Report;
	}
	if (g_underpathP) {
		g_underpath_lth = strlen(g_underpathP);
	} else if (g_mode >= Strip_source) {
		fprintf(stderr,"The -s,-w and -u option require the -p option\n");
		fprintf(stderr,"This is a safety measure\n");
		return(1);
	}
	find_dead(filenameP);
	printf("%d lines of code #ifdef'd out\n", g_lines_code);
	if (g_stripped) {
		printf("%d #ifdef _DEAD_ statements removed\n", g_stripped);
	}
	printf("%d functions are dead\n", g_dead_functions);
	if (g_failed) {
		printf("%d functions had no known boundaries\n", g_failed);
	}
	if (g_dead_lines) {
		printf("%d lines of source code marked dead\n", g_dead_lines);
	}
	if (g_updated) {
		printf("%d files updated\n", g_updated);
	}
	if (g_skipped) {
		printf("%d functions %d files not updated because not under %s\n", g_skipped, g_skipped_files, g_underpathP);
	}
	return 0;
}

